from unittest import mock
from agiletoolkit.test import gitrepo
from agiletoolkit.repo import RepoManager


def test_message():
    with gitrepo('slack'):
        m = RepoManager()
        with mock.patch('click.secho') as p:
            m.message('test')
            assert p.called



import os
import json

from dotenv import load_dotenv

load_dotenv()   # noqa

import click

from . import __version__
from .github import git
from .kong import kong

AGILE_CONFIG = os.environ.get('AGILE_CONFIG', 'agile.json')


@click.group(invoke_without_command=True)
@click.option(
    '--debug/--no-debug',
    is_flag=True,
    default=False,
    help='Run in debug mode'
)
@click.option(
    '--version',
    is_flag=True,
    default=False,
    help='Display version and exit'
)
@click.option(
    '--config', default=AGILE_CONFIG,
    type=click.Path(),
    help=f'Agile configuration json file location ({AGILE_CONFIG})'
)
@click.pass_context
def start(ctx, debug, version, config):
    """Commands for devops operations"""
    ctx.obj = {}
    ctx.DEBUG = debug
    if os.path.isfile(config):
        with open(config) as fp:
            agile = json.load(fp)
    else:
        agile = {}
    ctx.obj['agile'] = agile
    if version:
        click.echo(__version__)
        ctx.exit(0)
    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())


start.add_command(git)
start.add_command(kong)



import click

from ..utils import command
from ..repo import RepoManager


@click.command()
@click.pass_context
def remote(ctx):
    """Display repo github path
    """
    with command():
        m = RepoManager(ctx.obj['agile'])
        click.echo(m.github_repo().repo_path)



import os
import re
import json
import subprocess
from asyncio import get_event_loop
from contextlib import contextmanager

import click


PR_RE = re.compile('pr-*')
FORMAT = '%n'.join(['%H', '%aN', '%ae', '%cN', '%ce', '%s'])


class CommandError(click.ClickException):
    pass


def wait(awaitable):
    get_event_loop().run_until_complete(awaitable)


def deep_merge(source, destination):
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            deep_merge(value, node)
        else:
            destination[key] = value
    return destination


def sh(command, cwd=None, echo=None, env=None):
    if echo:
        echo(command)
    p = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        cwd=cwd,
        env=env,
        universal_newlines=True
    )
    p.wait()
    out, err = p.communicate()
    if p.returncode and err:
        raise CommandError(err)
    elif err:
        print(err)
    return out


def shi(command, cwd=None, echo=None, env=None):
    env = env or {}
    if echo:
        echo(command)
    p = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        cwd=cwd,
        env=dict(os.environ, **env),
        universal_newlines=True
    )
    while True:
        line = p.stdout.readline()
        if not line:
            break
        print(line.rstrip())
    p.wait()
    err = p.stderr.read()
    if p.returncode and err:
        raise CommandError(err)
    elif err:
        print(err)


@contextmanager
def command():
    try:
        yield
    except CommandError:
        raise


def error(exc):
    click.secho('ERROR: %s' % exc, err=True, fg='red', bold=True)


def niceJson(data):
    if isinstance(data, str):
        data = json.loads(data)
    return json.dumps(data, indent=4)


def semantic_version(tag):
    """Get a valid semantic version for tag
    """
    try:
        version = list(map(int, tag.split('.')))
        assert len(version) == 3
        return tuple(version)
    except Exception as exc:
        raise CommandError(
            'Could not parse "%s", please use '
            'MAJOR.MINOR.PATCH' % tag
        ) from exc


def gitrepo(root=None):
    if not root:
        cwd = root = os.getcwd()
    else:
        cwd = os.getcwd()
        if cwd != root:
            os.chdir(root)
    gitlog = sh('git --no-pager log -1 --pretty="format:%s"' % FORMAT,
                cwd=root).split('\n', 5)
    branch = sh('git symbolic-ref HEAD --short 2>/dev/null', cwd=root).strip()
    if not branch:
        branch = sh(
            "git branch -a --contains HEAD | sed -n 2p | awk '{ printf $1 }'"
        )
    if branch.startswith('remotes/origin/'):
        branch = branch[15:]
    try:
        current_tag = sh('git tag --points-at HEAD')
    except CommandError:
        current_tag = ''
    try:
        tag = sh('git describe --tags --abbrev=0')
    except CommandError:
        tag = ''
    tag = tag.strip()
    current_tag = current_tag.strip()
    remotes = [x.split() for x in
               filter(lambda x: x.endswith('(fetch)'),
                      sh('git remote -v', cwd=root).strip().splitlines())]
    if cwd != root:
        os.chdir(cwd)
    return {
        "head": {
            "id": gitlog[0],
            "author_name": gitlog[1],
            "author_email": gitlog[2],
            "committer_name": gitlog[3],
            "committer_email": gitlog[4],
            "message": gitlog[5].strip(),
        },
        "branch": branch,
        "current_tag": current_tag,
        "tag": tag,
        "pr": bool(PR_RE.match(branch)),
        "remotes": [{'name': remote[0], 'url': remote[1]}
                    for remote in remotes]
    }


def version():
    return sh('make version').strip().split()[-1]



import asyncio
import os
from typing import Dict

import yaml
from jinja2 import Template

from . import utils

NAMESPACES = {
    'local', 'development', 'dev', 'stage', 'prod', 'sandbox', 'production'
}


class Manager:

    def __init__(self, config=None, path=None, yes=False, namespace=None):
        self.config = config or {}
        self.path = path or os.getcwd()
        self.yes = yes
        self.message_brokers = []
        self.namespace = namespace
        self.init()

    def init(self) -> None:
        pass

    def get(self, name: str, DataClass: type) -> object:
        cfg = self.config.get(name) or {}
        return DataClass(**cfg)

    def message(self, msg: str) -> None:
        """Send a message to third party applications
        """
        for broker in self.message_brokers:
            try:
                broker(msg)
            except Exception as exc:
                utils.error(exc)

    def load_data(self, *paths: str, filename: str = None) -> Dict:
        filename = filename or self.filename(*paths)
        if not os.path.isfile(filename):
            raise utils.CommandError('%s file missing' % filename)
        with open(filename, 'r') as fp:
            data = yaml.load(fp, Loader=yaml.FullLoader) or {}
        data_namespace = data.pop(self.namespace, None)
        for namespace in NAMESPACES:
            data.pop(namespace, None)
        if data_namespace:
            data.update(data_namespace)
        data['namespace'] = self.namespace
        return data

    def manifest(self, values, *paths, filename: str = None) -> Dict:
        """Load a manifest file and apply template values
        """
        filename = filename or self.filename(*paths)
        with open(filename, 'r') as fp:
            template = Template(fp.read())
        return yaml.load(template.render(values), Loader=yaml.FullLoader)

    def filename(self, *paths) -> str:
        return os.path.join(self.path, *paths)

    def copy_env(self, *args, **kw):
        env = os.environ.copy()
        env.update(*args, **kw)
        return env

    def wait(self, coro):
        return asyncio.get_event_loop().run_until_complete(coro)



import os
import stat
import mimetypes
import logging
from urllib.parse import urlsplit

from .components import RepoComponents, GithubException
from ..utils import semantic_version


logger = logging.getLogger(__file__)

ONEMB = 2**20


class Assets(RepoComponents):
    pass


class Releases(RepoComponents):
    """Github repository endpoints
    """
    @property
    def assets(self):
        return Assets(self)

    def latest(self):
        """Get the latest release of this repo
        """
        url = '%s/latest' % self
        response = self.http.get(url, auth=self.auth)
        if response.status_code == 200:
            return response.json()
        elif response.status_code != 404:
            response.raise_for_status()

    def tag(self, tag):
        """Get a release by tag
        """
        url = '%s/tags/%s' % (self, tag)
        response = self.http.get(url, auth=self.auth)
        response.raise_for_status()
        return response.json()

    def delete(self, id=None, tag=None):
        if tag:
            assert not id, "provide either tag or id to delete"
            release = self.tag(tag)
            id = release['id']
        assert id, "id not given"
        return super().delete(id)

    def release_assets(self, release):
        """Assets for a given release
        """
        release = self.as_id(release)
        return self.get_list(url='%s/%s/assets' % (self, release))

    def upload(self, release, filename, content_type=None):
        """Upload a file to a release

        :param filename: filename to upload
        :param content_type: optional content type
        :return: json object from github
        """
        release = self.as_id(release)
        name = os.path.basename(filename)
        if not content_type:
            content_type, _ = mimetypes.guess_type(name)
        if not content_type:
            raise ValueError('content_type not known')
        inputs = {'name': name}
        url = '%s%s/%s/assets' % (self.uploads_url,
                                  urlsplit(self.api_url).path,
                                  release)
        info = os.stat(filename)
        size = info[stat.ST_SIZE]
        response = self.http.post(
            url, data=stream_upload(filename), auth=self.auth,
            params=inputs,
            headers={'content-type': content_type,
                     'content-length': str(size)})
        response.raise_for_status()
        return response.json()

    def validate_tag(self, tag_name, prefix=None):
        """Validate ``tag_name`` with the latest tag from github

        If ``tag_name`` is a valid candidate, return the latest tag from github
        """
        new_version = semantic_version(tag_name)
        current = self.latest()
        if current:
            tag_name = current['tag_name']
            if prefix:
                tag_name = tag_name[len(prefix):]
            tag_name = semantic_version(tag_name)
            if tag_name >= new_version:
                what = 'equal to' if tag_name == new_version else 'older than'
                raise GithubException(
                    'Your local version "%s" is %s '
                    'the current github version "%s".\n'
                    'Bump the local version to '
                    'continue.' %
                    (
                        str(new_version),
                        what,
                        str(tag_name)
                    )
                )
        return current


def stream_upload(filename):
    with open(filename, 'rb') as file:
        while True:
            chunk = file.read(ONEMB)
            if not chunk:
                break
            yield chunk



import click

from ..repo import RepoManager


@click.command()
@click.option(
    '--sandbox', is_flag=True,
    help='Validate only on sandbox/deploy branch', default=False)
@click.pass_context
def validate(ctx, sandbox):
    """Check if version of repository is semantic
    """
    m = RepoManager(ctx.obj['agile'])
    if not sandbox or m.can_release('sandbox'):
        click.echo(m.validate_version())



import os
import re
from dataclasses import dataclass
from urllib.parse import urlparse

from .manager import Manager
from .api import GithubApi, GithubException
from .slack import SlackIntegration
from . import utils


@dataclass
class Branches:
    dev: str = 'master'
    sandbox: str = 'deploy'


class RepoManager(Manager):

    def init(self):
        self.info = utils.gitrepo(self.path)
        SlackIntegration.add(self)

    def software_version(self):
        """Software version

        Requires the Makefile entry

        version:
            ....
        """
        return utils.version()

    def version(self):
        """Software version of the current repository
        """
        branches = self.branches()
        if self.info['branch'] == branches.sandbox:
            try:
                return self.software_version()
            except Exception as exc:
                raise utils.CommandError(
                    'Could not obtain repo version, do you have a makefile '
                    'with version entry?\n%s' % exc
                )
        else:
            branch = self.info['branch'].lower()
            branch = re.sub('[^a-z0-9_-]+', '-', branch)
            return f"{branch}-{self.info['head']['id'][:8]}"

    def validate_version(self, prefix='v'):
        """Validate version by checking if it is a valid semantic version
        and its value is higher than latest github tag
        """
        version = self.software_version()
        repo = self.github_repo()
        repo.releases.validate_tag(version, prefix)
        return version

    def skip_build(self):
        """Check if build should be skipped
        """
        skip_msg = self.config.get('skip', '[ci skip]')
        return (
            os.environ.get('CODEBUILD_BUILD_SUCCEEDING') == '0' or
            self.info['current_tag'] or
            skip_msg in self.info['head']['message']
        )

    def can_release(self, target=None):
        if self.skip_build() or self.info["pr"]:
            return False

        branch = self.info['branch']
        branches = self.branches()

        if branch not in (branches.dev, branches.sandbox):
            return False

        if target and branch != getattr(branches, target):
            return False

        if not target:
            target = 'sandbox' if branch == branches.sandbox else 'dev'

        if target == 'sandbox':
            try:
                self.validate_version()
            except GithubException:
                return False

        return True

    def github(self):
        return GithubApi()

    def branches(self):
        return self.get('branches', Branches)

    def target_from_branch(self):
        branches = self.branches()
        return 'dev' if self.info['branch'] == branches.dev else 'sandbox'

    def github_repo(self) -> str:
        url = self.info['remotes'][0]['url']
        if url.startswith('git@'):
            url = url.split(':')
            assert url[0] == 'git@github.com'
            path = url[1]
        else:
            path = urlparse(url).path[1:]
        bits = path.split('.')
        bits.pop()
        return self.github().repo('.'.join(bits))
